import logging
from numpy import array, ndarray

from . import scarf

"""
#### old method
def find_background(data, residual_baseline_func, baseline_asl_func):
    params = (array([0.01, 10 ** 5]))
    bounds = [array([0.001, 10 ** 5]), array([0.1, 10 ** 9])]
    baseline_values = least_squares(residual_baseline_func, params[:], args=(data.values,), bounds=bounds)
    p, lam = baseline_values['x']
    baseline_values = baseline_asl_func(data.values, lam, p, niter=10)
    return baseline_values

def residual_baseline(params, y):
    p, lam = params
    niter = 10
    baseline = baseline_als(y, lam, p, niter)
    residual = y - baseline
    return residual

def baseline_als(y, lam, p, niter=10):
    L = len(y)
    D = diags([1, -2, 1], [0, -1, -2], shape=(L, L - 2))
    w = ones(L)
    if niter < 1:
        raise ValueError("n iter is too small!")
    for i in range(niter):
        W = spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.transpose())
        z = spsolve(Z, w * y)
        w = p * (y > z) + (1 - p) * (y < z)
    return z
####
"""


def correct_negative_bg(y_bg_rm, bg):
    """
    function to shift the bg and the y data with the bg removed up if there are negative values.
    :param y_bg_rm: the y data with bg removed
    :param bg: the bg as an np array
    :return:
    """
    y_min = y_bg_rm.min()
    if y_min < 0:
        # then shift all the data up so no points are below zero
        y_bg_rm += abs(y_min)
        # and lower the bg we have calculated by that shift too
        bg -= abs(y_min)
    return y_bg_rm, bg


def background_finder(x_data, y_data, bg, scarf_params=False):
    """
    function to find the background given bg which is either a string or an np array specifying the type of background to remove.
    :param x_data: x data as np array
    :param y_data: y data as np array. npte x and y must be same length
    :param bg: an np array of size y_data or a string specifying the type of bg to remove
    :param scarf_params: the parameters for scarf background removal. if nothing is passed it defaults to none
    :return: bg as np array, y data with the bg removed, the parameters relating to the background
    """
    if bg == 'no':  # then user doesn't want to make a background
        logging.warning(
            "Warning: no background specified, so not using a background,"
            " this may prevent algorithm from converging")
        bg = array([0 for _ in y_data])  # set the background as 0 everywhere
        data_bg_rm_y = y_data  # no background subtracted
        params = 'no'

    elif bg == 'SCARF':
        logging.debug('using SCARF method for background subtraction')
        data_bg_rm_y, bg, params = scarf.perform_scarf(
            x_data, y_data, scarf_params)

    elif isinstance(bg, ndarray):
        assert len(bg) == len(y_data), \
            "the background generated or passed is of incorrect length"
        logging.debug('using numpy array as supplied by user')
        data_bg_rm_y = y_data - bg  # subtract user supplied background from the data
        params = 'user_specified'

    elif bg == 'OLD':
        logging.warning(
            'user specified old bg subtraction method which is no longer supported')
        # bg = find_background(y_data, residual_baseline, baseline_als) # find a background the old way
        # data_bg_rm_y = y_data - bg  # subtract background from the data
        # params = 'old_method'
        raise TypeError('old method is no longer supported')

    else:  # then it is the incorrect type
        logging.warning('user specified unknown bg subtraction method')
        raise TypeError(
            'the background passed is in the incorrect format, please pass as type np array')

    return bg, data_bg_rm_y, params



from . import utilities as utili
import logging
import matplotlib
import matplotlib.pyplot as plt
from numpy import argsort
from scipy.signal import find_peaks as peak_find
matplotlib.use('TkAgg')


def peak_finder(data, prominence, height=0, width=0):
    """
    function to find peaks given a prominence and data
    :param data: y_data as an np array
    :param prominence: height from top of base to tip as a float
    :param height: optional, the minimum height of any peaks found
    :param width: the minimum width of any peaks found
    :return: peak_info, a dict containing information of the found peaks details such as center indices,
    edge indices and amplitudes. Note all are sorted in order of amplitude
    """
    logging.debug(f'finding peaks based on prominence of {prominence}')
    # do a routine looping through until the right number of peaks is found

    # find the peak positions in the data
    peaks, properties = peak_find(
        data, prominence=prominence, height=height, width=width)

    peaks = list(peaks)  # convert to a list
    amps = list(properties['peak_heights'])  # store the heights
    # we will sort below in order of amplitudes
    sorted_indices = argsort(amps)[::-1]

    peak_info = {
        'center_indices': sort_lists(
            sorted_indices, peaks), 'right_edges': sort_lists(
            sorted_indices, list(
                properties['right_bases'])), 'left_edges': sort_lists(
                    sorted_indices, list(
                        properties['left_bases'])), 'amplitude': sort_lists(
                            sorted_indices, amps)}
    return peak_info


def sort_lists(sorted_indices, list_to_sort):
    """
    given a list of indices and a list to sort sort the list using the sorted_indices order
    :param sorted_indices: a list of indices in the order they should be e.g. [0,4,2,3]
    :param list_to_sort: the list which needs to be sorted in the indice order from sorted_indices
    :return: sorted list
    """
    return [list_to_sort[i] for i in sorted_indices]


def find_cents(prominence, y_data, find_all=False):
    """
    given a prominence and y_data find the peak_info which contains center indices. a redundant function which will be
    replace by peak_finder
    :param prominence:
    :param y_data:
    :param find_all:
    :return:
    """
    logging.debug(f'finding centre values based on prominence of {prominence}')
    peak_info = peak_finder(
        y_data,
        prominence,
        height=0,
        width=0)  # find the peak centers
    if find_all:
        return peak_info
    #center_indices = peak_info['center_indices']
    return peak_info


def auto_peak_finder(prominence, x_data, y_data):
    """
    automatic peak finding routine which finds peaks given a user supplied prominence, and presents the peaks until user
    is happy
    :param prominence: height from noise to tip of peaks
    :param x_data: x_data as np array
    :param y_data: y_data as np array
    :return: peak_info is a dict containing information about the peaks e.g. centers etc. also prominence used to find
    these peaks
    """
    while True:
        peak_info = find_cents(prominence, y_data, find_all=True)
        plt.plot(x_data, y_data)
        peak_coordinates = [x_data[ind] for ind in peak_info['center_indices']]
        for xc in peak_coordinates:
            plt.axvline(x=xc)
        print(
            f"Peak finder requires user input, please look at the following plot with prominence={prominence}")
        plt.show()
        ans = input(
            "If you are happy with the plot, type y. If not then please type a new prominence ")
        if ans == 'y':
            break
        else:
            try:
                prominence = float(ans)
            except ValueError:
                print("You entered an incorrect answer! Trying again...")
    plt.close()
    return peak_info, prominence


def user_peak_finder(x_data, y_data):
    """
    as the user to find the peak properties one by one, i.e. the centers, then amplitudes then sigmas.
    :param x_data: np array of x data
    :param y_data: np array of y data
    :return: peak_info dict containing data on peaks
    """
    peak_info = {'center_values': (), 'amplitude': (), 'sigma': ()}
    for key, peak_value in peak_info.items():
        while True:
            to_plot = peak_value
            if key == 'sigma':
                cents = peak_info['center_values']
                to_plot = [(cents[i] - width / 2, cents[i] + width / 2)
                           for i, width in enumerate(to_plot)]
            print(
                f"Peak finder requires user input, please look at the following plot of values for {key}: {peak_value}")
            plot_values(key, to_plot, x_data, y_data)
            ans = input(
                "If you are happy with the plot, type y. If not then please type a list of values seperated by commas ")
            if ans == 'y':
                break
            else:
                try:
                    peak_value_ = ans.split(',')
                    peak_value = [float(i) for i in peak_value_]
                except ValueError:
                    print("You entered an incorrect answer! Trying again...")
        plt.close()
        peak_info[key] = peak_value
    return peak_info


def plot_values(key, list_of_values, x_data, y_data):
    """
    plot the values of the data and any centers etc.
    :param key: the key of what is being plotted
    :param list_of_values: a lisst of the values that are being plotted e.g. center values
    :param x_data: an np array of x data
    :param y_data: an np array of y data
    :return:
    """
    plt.plot(x_data, y_data)
    if key == 'center_values':
        for xc in list_of_values:
            plt.axvline(x=xc)
    elif key == 'amplitude':
        for yc in list_of_values:
            plt.axhline(y=yc)
    elif key == 'sigma':  # these have to be passed in as band edges
        for band in list_of_values:
            plt.axvspan(band[0], band[1], alpha=0.1, color='red')
    plt.show()


def interactive_peakfinder(x_data, y_data, type='auto', prominence=1):
    """
    decide which interactive peak finder to use, i.e. the completely user guided one or the prominence based automatic one
    :param x_data: np array of x data
    :param y_data: np array of y data
    :param type: str: which type of peak finder to use
    :param prominence: optional, the prominence to pass to the auto peak finder
    :return: peak_info dict containing information of the peaks, and the prominence used
    """
    logging.debug('finding centres of peaks with user guided routine')
    if type == 'user':
        peak_info = user_peak_finder(x_data, y_data)
        prominence = None
    else:
        peak_info, prominence = auto_peak_finder(prominence, x_data, y_data)
        center_indices = peak_info['center_indices']
        peak_info['center_values'] = x_data[center_indices]
    return peak_info, prominence


def match_peak_centres(center_indices, y_data, prominence=1):
    """
    match the centers of peaks using some indices of guesses for the center and the actual data. uses a peak finder and
    finds the closest peak to each of the provided guesses
    :param center_indices: a list of indices of the centers
    :param y_data: np array of y data
    :param prominence: prominence used to find the centers. if this is too high then too few peaks will be found
    :return: a list of center indices
    """
    while True:
        peak_info_ = find_cents(prominence, y_data, find_all=True)
        center_indices_ = utili.find_closest_indices(
            peak_info_['center_indices'],
            center_indices)  # the indices might be slightly off so fix that
        if len(center_indices_) == len(center_indices):
            break
        elif len(center_indices_) > len(center_indices):
            center_indices_ = center_indices_[
                :len(center_indices)]  # as they're in order of prominence
            break
        else:
            # increase prominence until we get more than or equal to number.
            prominence *= 10
        # do something smarter!
    center_indices = [peak_info_['center_indices'][i] for i in center_indices_]
    return center_indices


def find_peak_details(x_data, y_data, type='auto', prominence=1):
    """
    given the data call the interactive_peakfinder which will launch a user guided routine to find peaks. then pull
    from this the results of that e.g. centers and widths etc
    :param x_data: np array of x data
    :param y_data: np array of y data
    :param type: type of user guided routine to use
    :param prominence: prominence to pass to user guided routine
    :return: int number of peaks being used, peak info dictionary, peak_info_dict containing information of the peaks,
    prominence used to find peaks
    """
    logging.debug(
        f'finding peak details based on prominence of {prominence}, and user provided details:'
        f'no_peak')

    peak_info, prominence = interactive_peakfinder(
        x_data, y_data, type, prominence)
    no_peaks = max(len(value) for value in peak_info.values())
    center = peak_info['center_values']
    amplitude = peak_info['amplitude']
    sigma = peak_info.get('sigma', ())
    for i, sig in enumerate(sigma):
        x1 = x_data[0] + sig  # this is a bit dodgy??
        # this finds the closest x value to the width plus the first
        sigma[i] = utili.find_closest_indices(x_data, x1)
        # x value. should really do this from the raw widths
    type = ["LorentzianModel" for i in range(no_peaks)]
    peak_info_dict = {}

    peak_info_dict['center'], peak_info_dict['amplitude'], peak_info_dict['sigma'], \
        peak_info_dict['type'] = center, amplitude, sigma, type

    return no_peaks, peak_info_dict, prominence


def make_bounds(x_data, y_data, no_peaks, peak_info_dict):
    """
    user guided routine to find peak parameter bounds given inputs
    :param x_data: np array of x data
    :param y_data: np array of y data
    :param no_peaks: no of peaks as int
    :param peak_info_dict: dictionary containing all the peak information
    :return: bounds dictionary containing user defined bounds on all the keys in peak_info_dict
    """
    logging.debug(f'making bounds based on: no_peaks:{no_peaks}, '
                  f'peak_info_dict: {peak_info_dict}')
    bounds = {}
    for key in peak_info_dict:
        while True:
            ans = input(f"Do you want to create bounds for {key}? \n"
                        f"Current values are: {peak_info_dict[key]} \n"
                        f"type y or n")
            if ans == 'y':
                bounds[key] = interactive_bounds(
                    x_data, y_data, len(peak_info_dict[key]))
                break
            elif ans == 'n':
                break
            else:
                print("incorrect answer, please type y or n")
    return bounds


def interactive_bounds(x_data, y_data, no_bounds: int):
    """
    interactive bound find for the data
    :param x_data: np array of x data
    :param y_data: np array of y data
    :param no_bounds: how many bounds to create.
    :return: bounds as a list of values for whatever is being calculated.
    """
    bounds = [(0, 0) for i in range(no_bounds)]
    while True:
        plt.plot(x_data, y_data)
        for bound in bounds:
            plt.axvspan(bound[0], bound[1], alpha=0.1, color='red')
        print(f"current bounds are: {bounds}")
        plt.show()
        ans = input(
            f"Please enter a list of values for the bounds, comma delimited. You must enter {2*no_bounds}"
            f" numbers. type y if you are happy with these bounds")
        if ans == 'y':
            break
        else:
            try:
                bounds = ans.split(',')
                bounds = [float(i) for i in bounds]
                if len(bounds) != 2 * no_bounds:
                    raise ValueError("incorrect no of bounds passed")
                else:
                    bounds = [(bounds[2 * i], bounds[2 * i + 1])
                              for i in range(no_bounds)]
            except Exception as e:
                print("You entered an incorrect answer! Trying again...")
    plt.close()
    return bounds



import logging
from ast import literal_eval
from os.path import basename
from os.path import join
from os import rename

from . import map_scan_tools
from . import multi_obj
from . import parsing
from . import utilities as utili


def main():
    """
    the script to run a mapscan analysis for a user based on a parameters file or manually passed args
    :return:
    """

    args = parsing.parse_user_args()

    arguments = parsing.using_user_args(args)

    try:  # a user can pass in a list of filenames or just one
        file_name = basename(literal_eval(arguments['datapath'])[0])
        log_name = file_name
    except (SyntaxError, ValueError):  # assume its just a string and not a list passed
        file_name = arguments['datapath']
        log_name = arguments['datapath']
        arguments['datapath'] = [
            f"{arguments['datapath']}",
        ]  # as this is what multiobj needs

    # setup logger
    log_filename = utili.setup_logger(log_name.split('.')[0]) # don't name it with the .txt

    logging.info('creating multi_obj object')
    bag = multi_obj.main(arguments)  # create a Thunder object

    # get the first thunder object
    bag.first = next(iter((bag.thunder_bag.keys())))

    if (arguments.get('clip_data',
                      False) and not arguments.get('clips',
                                                   False)) or arguments.get('bg_first_only',
                                                                            False) or arguments.get('peakf_first_only',
                                                                                                    False) or arguments.get('bounds_first_only',
                                                                                                                            False):  # then we want the user to decide which thunder object to use
        logging.info('choosing spectrum for data')
        # choose which spectrum to base everything off of if user wants to use
        # one spectra to
        bag.choose_spectrum()
        # choose parameters
        logging.info(
            f'using spectra {bag.first} as spectra to set any user variables from')

    # clip the data if weird edges
    if arguments.get('clip_data', False):
        logging.info('clipping data')
        bag.clip_data(arguments.get('clips', False))

    # cosmic ray removal goes here
    ##########################################################################

    # remove background
    if arguments.get('bg_first_only', False):
        logging.info(
            'determining background conditions for all based on user guided for first')
        bag.bg_param_setter()
    logging.info('removing background from data for all thunder objects')
    bag.remove_backgrounds()

    # normalisation
    if arguments.get('normalise', False):
        logging.info('normalising data using svn normalisation')
        bag.normalise_data()
        # then shift the values up since normalisation may create negative
        # values
        for thund in bag.thunder_bag.values():
            thund.y_data_bg_rm = thund.y_data_bg_rm - min(thund.y_data_bg_rm)

    # find peaks
    if arguments.get('find_peaks', False):
        logging.info('setting peak information for all thunder objects')
        print("Warning: Finding peaks automatically will overwirte and peak_info_dict supplied")
        logging.info(
            'running user guided routine to determine peak information')
        bag.find_peaks()
    elif arguments.get('adj_params', False):
        logging.info('setting peak information for all thunder objects')
        bag.peaks_adj_params()

    # find bounds
    if arguments.get('find_bounds', False):
        logging.info('finding bounds via user guided routine')
        bag.bound_setter()
    else:
        logging.info(
            "setting all bounds to either user supplied or preset: {'amplitude': False, 'center': False, 'sigma': False}")
        # should really do this in the thunderobj # also don't set as these
        # presets!
        bounds = arguments.get('bounds', {})
        bag.bound_setter(bounds)

    # fit peaks
    logging.info('fitting peaks for all')
    bag.fit_peaks()

    # fit params dictionary
    # store all the peak parameters in a dictionary, so the keys are e.g. sigma, center, amplitude, and the values are
    # dictionaries with keys as the run number with values as lists of values for all the peaks for that run
    # this will for now assume the same types of peak for all fits!
    logging.info('making fit parameters dictionary')
    bag.make_fit_params()

    # fetch stats etc
    logging.info('making stats dictionary')
    bag.get_fit_stats()

    # make directory to save everything in
    file_name, dirname = parsing.make_user_files(arguments, file_name)

    # plot map scan
    logging.info('plotting map scans')
    bag.make_map_matrices()  # make the mapscan arrays
    map_scan_tools.plot_map_scan(
        getattr(
            bag,
            'fit_params'),
        bag.map_matrices,
        bag.X_coords,
        bag.Y_coords,
        dirname)

    # save individual plots for each of the failed fits
    logging.info('saving failed fit plots')
    bag.save_failed_plots(dirname)

    # save a gif of every single plot
    if arguments.get('make_gif', False):
        logging.info('saving all plots as gif')
        utili.gif_maker(bag.thunder_bag, f'{dirname}/all_data.gif')

    # put here some code for cluster analysis and pca
    logging.info('not currently doing cluster analysis or pca')
    ##########################################################################

    # save the bag object and it reports
    logging.info('saving fit reports on stats and fitting parameters')
    utili.save_fit_report(
        getattr(
            bag,
            'stats'),
        path=dirname,
        filename=f"report.json")
    utili.save_fit_report(
        getattr(
            bag,
            'fit_params'),
        path=dirname,
        filename=f"peak_info.json")
    logging.info('saving thunderbag object')
    utili.save_thunder(bag, path=dirname, filename=f"bag_object.d")

    # move the log file in with all the rest of it
    log_filename_ = str(join(dirname, f'{file_name}.log'))
    # use os.rename to move the log file to the final destination
    rename(log_filename, log_filename_)
    utili.save_fit_report(
        arguments,
        path=dirname,
        filename=f"inpargs.json")



from . import utilities as utili
import logging
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib
import matplotlib.pyplot as plt
from numpy import unique, round, array, nanmin, nanmax, nan, nanpercentile, nanmean
from scipy.sparse import coo_matrix

matplotlib.use('TkAgg')


# funcs for plotting
def shift_map_matr(coordinates_array):
    """
    function to shift all the coordinates in a coordinates array so that it start from zero.
    :param coordinates_array: an np array of mxn with the coordinates of each scan at each position
    :return: the shifted coordinates array
    """
    logging.debug('shifting coordinates array')
    coordinates_array[:, 0] = coordinates_array[:, 0] - \
        min(coordinates_array[:, 0])
    coordinates_array[:, 1] = coordinates_array[:, 1] - \
        min(coordinates_array[:, 1])
    return coordinates_array


def generate_map_matrix(coordinates, peak_label_vals):
    """
    function to generate a map matrix based on a dictionary of coordinates and peak_label_vals. both have matching keys
    and the coordinates values are (x,y) tuples, peak_label_vals is a dictionary of values of what is being mapped.
    :param coordinates: dictionary of keys corresponding to thunder objects, values are (x,y)
    :param peak_label_vals: same keys as coordinates but with values which will be mapped to a matrix
    :return: a matrix of dimensions from coordinates. each
    element corresponds to a new scan, so assumes a uniform grid. the spacing of scans isn't coded in here, but can be
    accessed from the X and Y which are returned (lists of the coodinates)
    """
    X = []  # initialise these
    Y = []
    Z = []

    for key in peak_label_vals.keys():  # corresponding to a pixel
        x, y = coordinates[key]  # coordinates
        z = peak_label_vals[key]  # value at those coordinates
        if isinstance(z, str):  # not sure why this is here.
            return [], []  # probably will break if this happens. better to raise an error
        X.append(x)  # append to coordinates and values to the lists
        Y.append(y)
        Z.append(z)
    x_step = unique(X)[1] - unique(X)[0]  # what are the step sizes in x
    # turn the x coordinates into an array of integer steps
    xx = (round(X / x_step)).astype(int)
    y_step = unique(Y)[1] - unique(Y)[0]
    yy = round((Y / y_step)).astype(int)
    # build a dense array by using sparse array notation and converting
    data_ = coo_matrix((Z, (yy, xx))).toarray()
    # if its zero we assume thats because there was no data taken there.
    data_[data_ == 0] = nan

    return X, Y, data_


def map_scan_matrices_from_dicts(coordinates, values):
    """
    function to generate dictionaries of X and Y coordiantes, as well as dict of data matrices. does this for all
    the peak types in the passed in dictionaries. keys are the properties e.g. 'center' etc
    :param coordinates: dictionary of coordinates for each property (usually doesn't differ though so is a repeat of the same ones)
    :param values: dict of values for each property (key). e.g. 'center' has value which is a dict of run keys and values at that key
    :return: data dict, and X_, Y_ coordinate dicts. same keys as input
    """
    logging.debug('generating map matrices')
    peak_labels = list(values.values())[0].keys()
    data = {}
    X_ = {}
    Y_ = {}
    for peak_label in peak_labels:
        peak_label_vals = {
            key: values[key].get(
                peak_label,
                nan) for key in values.keys()}
        X, Y, data_ = generate_map_matrix(coordinates, peak_label_vals)
        data[peak_label] = data_
        X_[peak_label] = X
        Y_[peak_label] = Y
    return data, X_, Y_


def map_plotter(data, X, Y):
    """
    function to plot a mapscan given a data matrix and the actual coordinates to plot on the map
    :param data: np array of values
    :param X: x coordinates corresponding to each element in data in x direction
    :param Y: same as for x
    :return: figure and axis to be reused/saved etc later
    """
    f = plt.figure()
    ax = plt.gca()
    magma_cmap = matplotlib.cm.get_cmap('magma')
    magma_cmap.set_bad(color='green')
    im = plt.imshow(
        data,
        cmap=magma_cmap,
        vmin=nanpercentile(
            data,
            3),
        extent=[
            min(X),
            max(X),
            max(Y),
            min(Y)],
        vmax=nanpercentile(
            data,
            97))  # we plot the 99th percentile and 1st percentil as the max and min
    # colours
    plt.xlabel('x coordinates')
    plt.ylabel('y coordinates')
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax)

    return f, ax


def map_scan_plot_dicts(data_mat, X_coords, Y_coords):
    """
    plot all the data and save all the plots into dicts which are returned
    :param data_mat: data matrices as dicts. keys are for the properties for which each map matrix is the value
    :param X_coords: x coordinates as dicts. keys same as data_mat
    :param Y_coords: same as for x
    :return:
    """
    logging.debug('plotting mapscans')
    peak_labels = data_mat.keys()
    figs = {}
    axs = {}
    for peak_label in peak_labels:
        data = data_mat[peak_label]
        X = X_coords[peak_label]
        Y = Y_coords[peak_label]
        f, ax = map_plotter(data, X, Y)
        figs[peak_label] = f
        axs[peak_label] = ax
    return figs, axs


def get_cent_mean(peak_label, fit_params):
    """
    function to return the mean center values from fit_params dict. isn't a great abstraction barrier though!
    :param peak_label: the label of the peak you want the mean center value for e.g. 'm0' for mdoel 0
    :param fit_params: fit params dict with keys of properties for peaks e.g. 'center' and 'amplitude'
     these have values which are dicts which include peak_label as a key. those values are the values for that property
     for each peak
    :return: the mean center value
    """
    cents = [
        cent.get(
            peak_label,
            0) for cent in fit_params.get(
            'center',
            nan).values()]
    cent_mean = nanmean(cents)
    return cent_mean


def save_mapscan(peak_label, fit_params, plot, dirname, p):
    """
    given a mapscan plot save it and add a title
    :param peak_label: the label of the peak the mapscan is for
    :param fit_params: the dicitonary of all peak properties and peaks values
    :param plot: the plot to save
    :param dirname: str: where to save it
    :param p: str: what is the peak being saved
    :return:
    """
    cent_mean = get_cent_mean(peak_label, fit_params)
    plot[peak_label].suptitle(
        f"{p}_{peak_label}_heatmap. peak {peak_label} is centered at:  {cent_mean}")
    utili.save_plot(
        plot[peak_label],
        path=dirname,
        figname=f"{p}_{peak_label}.svg")


def mapscans_for_parameter(
        map_matrices,
        X_coords,
        Y_coords,
        p,
        fit_params,
        dirname):
    """
    given the map matrices coordinates dictionaries and fit parameters, plot all the mapscans and then call them all to be saved
    :param map_matrices: a dict of map matrices, keys are the type of map to save e.g. 'center'. the values are dicts of
    actual data matrices as values and keys correspond to the model keys e.g. 'm0'
    :param X_coords: coordinates dict with keys same as map_matrices
    :param Y_coords: same as for x
    :param p: key for what mapscan to access
    :param fit_params: fit parameters dictionary as made by multi_obj method
    :param dirname: where to save the mapscans
    :return:
    """
    data_mat = map_matrices[p]
    plot, ax = map_scan_plot_dicts(data_mat, X_coords[p], Y_coords[p])
    for peak_label in plot.keys():
        save_mapscan(peak_label, fit_params, plot, dirname, p)
    plt.close()


def plot_map_scan(fit_params, map_matrices, X_coords, Y_coords, dirname):
    """
    function to plot the mapscans given user input on what to plot for.
    :param fit_params: a dictionary of dicts. 1st keys are properties e.g. 'center' and second keys are for each model
    e.g. 'm0'. values are a dict of the actual values for each pixel in the map
    :param map_matrices: same as fit_params but the values are map matrices instead of dicts
    :param X_coords: same as map_matrices except the values are lists of X_coords
    :param Y_coords: same as x
    :param dirname: str: where to save the plots
    :return:
    """
    logging.debug(
        'runnning user input routine to generate/save user chosen variables in maps')
    while True:
        ans = input(
            "making map scans, please input which property you would like to scan. options are:"
            f"\n {[p_ for p_ in fit_params.keys()]}, or type 'all' to plot all, or type e to exit")
        if ans == 'e':
            break
        elif ans == 'all':
            for p in fit_params.keys():
                mapscans_for_parameter(
                    map_matrices, X_coords, Y_coords, p, fit_params, dirname)
            break  # break the while loop
        else:
            try:
                p = ans
                mapscans_for_parameter(
                    map_matrices, X_coords, Y_coords, p, fit_params, dirname)
            except KeyError:
                print('wrong answer entered, trying again!')
#
    return map_matrices



import logging
from copy import deepcopy
from difflib import get_close_matches
from re import findall
from typing import Dict, Union

import matplotlib.pyplot as plt
from lmfit.model import ModelResult
from numpy import ndarray

from . import plotting
from . import utilities as utili
from .background import background_removal as bg_remove
from . import peak_finding
from . import peak_fitting

# TODO: need to fail if peak fitting doesn't work!


class Thunder():
    """
    thunder object with all the methods we love inside it. Name generated using WuTang Clan name generator.
    """

    def __init__(self, input, x_data=None, y_data=None, e_data=None):
        """
        initialise all the attributes which will be set later
        :param input: a dictionary of all user inputs to be set
        :param x_data: the x_data np array
        :param y_data: y data np array
        :param e_data: error dat anp array
        """
        self.input: Union[Thunder, Dict] = input

        self.x_ind: int = 0
        self.y_ind: int = 1
        self.e_ind: Union[int, None] = None

        self.x_data: Union[None, ndarray] = x_data
        self.y_data: Union[None, ndarray] = y_data
        self.e_data: Union[None, ndarray] = e_data

        self.y_data_bg_rm = None
        self.y_data_norm = None

        self.datapath: str = './data.txt'

        self.no_peaks: int = 0
        self.background: str = "SCARF"  # fix this
        self.scarf_params: Union[None, Dict] = None

        self.peak_info_dict: Dict = {}
        self.peak_finder_type: str = 'auto'

        self.bounds: Dict = {}

        self.peaks: ModelResult
        self.plot: plt = None
        self.fit_report: {} = {}
        self.peak_params: {} = {}
        self.model = None  # give type!

        self.free_params: int = 0
        self.p_value: int = 0
        self.chisq: int = 0

        self.method: str = 'leastsq'
        self.tol: float = 0.0000001

        if isinstance(
                input,
                Thunder):  # if only pass one but its already a thunder object then just use that
            # add all the details in depending on args
            self.overwrite_thunder(input)
        elif isinstance(input, dict):  # if its a dict then we pull all the values from it
            # add all the details in depending on args
            self.create_thunder(input)
        else:
            raise TypeError('Cannot convert input to Thunder object')

        if isinstance(
                self.x_data,
                ndarray) and isinstance(
                self.y_data,
                ndarray):
            pass  # they're already loaded as they've been passed
        else:
            self.x_data, self.y_data, self.e_data = utili.load_data(
                self.datapath, self.x_ind, self.y_ind)  # load the data

    def overwrite_thunder(self, inp):
        """
        Methods for creating a thunder object from another, pull everything out from it and set it as our attributes here
        :param inp: a thunder object
        :return:
        """
        logging.debug('overwriting thund obj')
        thun = inp

        if thun.x_data and thun.y_data:
            self.x_data = thun.x_data
            self.y_data = thun.y_data
        else:
            self.x_ind = thun.x_ind
            self.y_ind = thun.y_ind
            self.e_ind = thun.e_ind
            self.datapath = thun.datapath

        if thun.y_data_bg_rm:
            self.y_data_bg_rm = thun.y_data_bg_rm
        if thun.y_data_norm:
            self.y_data_norm = thun.y_data_norm

        self.no_peaks = thun.no_peaks
        self.background = thun.background
        self.scarf_params = thun.scarf_params

        self.peak_info_dict = thun.peak_info_dict
        self.peak_finder_type = thun.peak_finder_type

        self.bounds = thun.bounds

        self.method = thun.method
        self.tol = thun.tol

    def create_thunder(self, inp: Dict):
        """
        Used to create a thunder object given different input types
        :inp: a dictionary containing attributes we will asign to the current thunder obj
        :return: None, we modify the object unless a spec1d object is passed, in which case we return that
        """
        logging.debug('creating thund obj')
        try:  # if e_ind is missing we don't reallt care
            self.e_ind = inp['e_ind']
        except KeyError as e:
            logging.info(
                f"KeyError: Missing field in the data dictionary: {e}")

        try:  # If of these are missing then need to fail here
            self.datapath = inp['datapath']
            self.x_ind = inp['x_ind']
            self.y_ind = inp['y_ind']
        except KeyError as e:
            raise KeyError(f"Missing vital information to load object: {e}")

        self.no_peaks = inp.get('no_peaks', self.no_peaks)
        self.background = inp.get('background', self.background)
        # todo: do some check on background here to set it to an np array

        self.scarf_params = inp.get('scarf_params', self.scarf_params)

        self.peak_info_dict = inp.get('peak_info_dict', self.peak_info_dict)
        self.peak_finder_type = inp.get(
            'peak_finder_type', self.peak_finder_type)

        self.bounds = inp.get('bounds', self.bounds)
        self.method = inp.get('method', self.method)
        self.tol = inp.get('tol', self.tol)

    def clip_data(self, clips=None):
        """
        method to clip the data. if clips passed then will not be interactive
        :param clips: if a two element list of numbers then will use, otherwise will interactively get clips
        :return:
        """
        clip_left, clip_right = utili.clip_data(
            getattr(
                self, 'x_data'), getattr(
                self, 'y_data'), clips)
        # the cli_data func will return the INDICES of the data to clip so we
        # can just do it below
        setattr(self, 'x_data', getattr(self, 'x_data')[
                clip_left:clip_right])  # clip the data points
        setattr(self, 'y_data', getattr(self, 'y_data')[clip_left:clip_right])

    def remove_bg(self):
        """
        call the background_finder function and then set attributes from it
        :return:
        """
        background, y_data_bg_rm, params = bg_remove.background_finder(
            getattr(
                self, 'x_data'), getattr(
                self, 'y_data'), getattr(
                self, 'background'), getattr(
                    self, 'scarf_params'))
        setattr(self, 'background', background)
        setattr(self, 'y_data_bg_rm', y_data_bg_rm)
        setattr(self, 'params', params)

    def normalise(self):
        """
        normalise the data by calling the appropriate func and then setting attributes
        :return:
        """
        y_data_bg_rm, background, y_data_norm = utili.normalise_all(
            'y_data_bg_rm', 'background', 'y_data')
        setattr(self, 'background', background)
        setattr(self, 'y_data_bg_rm', y_data_bg_rm)
        setattr(self, 'y_data_norm', y_data_norm)

    def find_peaks(self):
        """
        find peaks interactively then set the peak positions. currently this relies on peaks with the center att
        :return:
        """
        no_peaks, peak_info_dict, prominence = peak_finding.find_peak_details(
            getattr(
                self, 'x_data'), getattr(
                self, 'y_data_bg_rm'), getattr(
                self, 'no_peaks'), getattr(
                    self, 'peak_finder_type', 'auto'))
        setattr(self, 'no_peaks', no_peaks)  # set values
        center_indices = utili.find_closest_indices(
            self.x_data, peak_info_dict['center'])  # get the indices from the x centres
        center_indices = peak_finding.match_peak_centres(
            center_indices, self.y_data)  # match to the peakfinding cents
        peak_centres = self.x_data[center_indices]  # convert back to x values
        peak_info_dict['center'] = peak_centres  # set values
        setattr(self, peak_info_dict, peak_info_dict)

    def bound_setter(self, bounds=None):
        """
        set the bounds for the attributes. this is a dictionary which will be used later
        :param bounds:
        :return:
        """
        if not bounds:
            bounds = peak_finding.make_bounds(
                getattr(
                    self, 'x_data'), getattr(
                    self, 'y_data'), getattr(
                    self, 'no_peaks'), self.peak_info_dict)
        setattr(self, 'bounds', bounds)  # set values

    def fit_peaks(self):
        """
        call the peak fitting routine
        :return:
        """
        specs, model, peak_params, peaks = peak_fitting.fit_peaks(
            getattr(
                self, 'x_data'), getattr(
                self, 'y_data_bg_rm'), getattr(
                self, 'peak_info_dict'), getattr(
                    self, 'bounds'), getattr(
                        self, 'method'), getattr(
                            self, 'tol'))
        setattr(self, 'specs', specs)
        setattr(self, 'model', model)
        setattr(self, 'peak_params', peak_params)
        setattr(self, 'peaks', peaks)

    # plot_all and fit_report need imporovements e.g. to check which
    # attributes exists in the object
    def plot_all(self, ax=None, plot_unc=True):
        """
        plot all the peaks, background, original and bg subtracted data and uncertainty if plot_unc is True. Will save
        this plot and self.plot which can be accessed later to either save or show the plots
        :param ax: if you want to plot on an existing plot then pass the matplotlib ax obj
        :param plot_unc: bool, if true will plot the uncertainties, if not then it won't
        :return:
        """
        logging.debug('plotting all for thund obj')
        # plot each component of the model
        ax, plt, fig = plotting.plot_fits(
            self.x_data, self.peaks.eval_components(), ax=ax)
        ax, plt, fig = plotting.plot_background(
            self.x_data, self.background, ax=ax, fig=fig)  # plot the background supplied by user
        try:
            if plot_unc:
                ax, plt, fig = plotting.plot_uncertainty_curve(
                    self.x_data, self.peaks.eval_uncertainty(
                        sigma=3), self.peaks.best_fit, ax=ax, fig=fig)  # plot a band of uncertainty
        except TypeError:
            logging.warning(
                'There are not uncertainties available because the bounds are too restrictive.'
                'The covariance matrix is not valid near the bounds so widen the bounds for whatever is '
                'causing the issue')
        ax, plt, fig = plotting.plot_data(
            self.x_data, self.y_data, ax=ax, fig=fig)  # plot the raw data
        ax, plt, fig = plotting.plot_fit_sum(
            self.x_data, self.peaks.best_fit, 0, ax=ax, line='k--', fig=fig)  # plot the fitted data
        ax, plt, fig = plotting.plot_data(
            self.x_data, self.y_data_bg_rm, ax=ax, line='g--', fig=fig)  # plot the raw data bg rm

        ax.minorticks_on()
        ax.grid(which='minor', alpha=0.2)
        ax.grid(which='major', alpha=0.5)

        self.plot = plt

        return ax, fig

    def gen_fit_report(self):
        """
        generate a fit report which will be a dicitonary containing useful info such as chi sq etc and the errors
        on each of the model attributes
        :return:
        """
        logging.debug('genertaing fit report for thund obj')
        self.fit_report = {mod_no: {}
                           for mod_no in range(len(self.peak_info_dict['type']))}

        # total fit data
        self.fit_report['chi_sq'] = self.chi_sq
        self.fit_report['free_params'] = self.free_params
        self.fit_report['p_value'] = 'not implemented'

        for parameter, param_obj in self.peaks.params.items():
            model_no = int(findall(r'\d+', parameter)[0])
            param_type = parameter.split('__')[1]

            if param_type:
                value = param_obj.value
                err = param_obj.stderr
                type = utili.safe_list_get(
                    self.peak_info_dict.get(
                        'type', [
                            False, ]), model_no, False)
                bounds = utili.safe_list_get(
                    self.bounds.get(
                        f'{param_type}', [
                            False, ]), model_no, False)

                fit_info = {"value": value,
                            "stderr": err,
                            "peak_type": type,
                            "bounds": bounds}

                self.fit_report[model_no][param_type] = fit_info


def main(arguments):
    """
    if you call this then it will create and return the thunder obj for you
    :param arguments: a thunder object or a dicitonary to initialise the thunder obj
    :return:
    """
    thunder = Thunder(deepcopy(arguments))  # load object
    return thunder
